/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                        LibUnixMG2.c                     **/
/**                                                         **/
/** This file contains implementation for some commonly     **/
/** used NetBSD/hpcmips routines.                           **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**               Hiroyuki Yanai  2000                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999, 2000
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 *      Shin Takemura (takemura@netbsd.org)
 *      SATO Kazumi (sato@netbsd.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef UNIX

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>

/* --- MG2 --- */
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <termios.h>
#include <fcntl.h>
#include <dev/wscons/wsconsio.h>
#include <dev/wscons/wsdisplay_usl_io.h>
#include <dev/hpc/hpcfbio.h>

#include "LibUnixMG2.h"

#define DEF_ROWBYTE2    256     /* XX MC-R300 */
#define DEF_ROWBYTE8    1024    /* XX MC-R500 */
#define DEF_ROWBYTE16   1600    /* XX MC-R510 */
#define FBDEV	 "/dev/ttyE0"
#define MOUSEDEV "/dev/wsmouse0"

#define KEYBUF_SIZE	2048

static int mouse_fd = -1;
static int mouse_x = 0;
static int mouse_y = 0;

static int fb_fd = -1;
static int fb_off = 0;
static int depth;
static char *platform = NULL;
static int rowbytes;
static char *fb = NULL;
static int width, height;
static u_char red[256], green[256], blue[256];
static struct termios save_tty, cur_tty;
static int key_buf[KEYBUF_SIZE];
static int key_buf_cnt = 0;
/* --- --- */


#define NEXT(N) ((N+1)%MAXALLOC)  /* Used when allocating colors */

struct Timer
{
  struct Timer *Next;
  int Freq;
  void (*Handler)(void);
};


struct Timer *FirstTimer = NULL;
static int TimerLock     = 0;

static void MasterTimer(int Arg);


struct _platinfo {
  int depth;
  int rowbyte;
  int off;
  char name[20];
} platinfo[] = {
  { 2, 80, 0,  "FreeStyle" },
  { 2, 160, 0,  "MC-R32" },
  { 2, 256, 0,  "MC-R3" },
  { 2, 256, 0,  "MC-CS" },
  { 2, 256, 0,  "MobileGearII for Do" },
  { 2, 256, 0,  "Cassiopeia E55" },
  { 2, 256, 0,  "Cassiopeia for DoCo" },
  { 2, 256, 0,  "E-55" },
  { 8, 1024, 0,  "MC-R500" },
  { 8, 1024, 0,  "MC-R510" },
  { 8, 640, 0,  "INTERTOP" },
  { 16, 1280, 256, "MC-R530" }, 
  { 16, 1280, 256, "MC-R430" }, 
  { 16, 1600, 0, "MC-R5" }, 
  { 16, 1600, 1536, "MC-R730" }, 
  { 16, 1600, 0, "MC-R700" }, 
  { 16, 512, 0,  "Cassiopeia E100" },
  { 16, 512, 0,  "E-100" },
  { 16, 512, 0,  "Cassiopeia E500" },
  { 16, 512, 0,  "E-500" },
  { 16, 1280, 0,  "WorkPad" },
  { 16, 1280, 0,  "z50" },
  { 0, 0, 0, 0 }
};

int open_mouse(void)
{
  if((mouse_fd = open(MOUSEDEV, O_RDWR )) < 0)
    return -1;

  return 0;
}

void close_mouse(void)
{
  if(mouse_fd == -1)  return;
  close(mouse_fd);
  mouse_fd = -1;
}

void mouse_event(void)
{
  struct wscons_event ev[16];
  fd_set fds;
  struct timeval to;
  int i, n, r;

  if (mouse_fd < 0) return;

retry:
  FD_ZERO(&fds);
  FD_SET(mouse_fd,&fds);
  to.tv_sec = 0;
  to.tv_usec = 0;
  if (select(mouse_fd+1,&fds,0,0,&to) != 1) return;
    r = read(mouse_fd,ev,sizeof(ev));
    if (r < 0)
    {
      close(mouse_fd);
       mouse_fd = -1;
       return;
    }
    n = r/sizeof(ev[0]);
//printf("event count = %d \n",n);
    for (i=0; i<n; i++)
    {
      switch(ev[i].type)
      {
        case WSCONS_EVENT_MOUSE_DELTA_X:
          mouse_x += ev[i].value;
//fprintf(stderr, "MOUSE_DELTAX %d\n",mouse_x);
          break;

        case WSCONS_EVENT_MOUSE_DELTA_Y:
          mouse_y += ev[i].value;
//fprintf(stderr, "MOUSE_DELTAY %d\n",mouse_y);
          break;

        case WSCONS_EVENT_MOUSE_ABSOLUTE_X:
          mouse_x = ev[i].value;
//fprintf(stderr, "MOUSE_ABSX %d\n",mouse_x);
          break;

        case WSCONS_EVENT_MOUSE_ABSOLUTE_Y:
          mouse_y = ev[i].value;
//fprintf(stderr, "MOUSE_ABSY %d\n",mouse_y);
          break;
      }
    }
}

void get_mouse_values(int *x, int *y, int *btn)
{
  *x = mouse_x;
  *y = mouse_y;
  *btn = 0;
}


static void text_mode(void)
{
  int dispmode;

  dispmode = WSDISPLAYIO_MODE_EMUL;
  if (ioctl(fb_fd, WSDISPLAYIO_SMODE, &dispmode) < 0) {
    perror("ioctl(WSDISPLAYIO_SMODE) WSDISPLAYIO_MODE_EMUL");
  }
}

static void graph_mode(void)
{
  int dispmode;

  dispmode = WSDISPLAYIO_MODE_MAPPED;
  if (ioctl(fb_fd, WSDISPLAYIO_SMODE, &dispmode) < 0) {
    perror("ioctl(WSDISPLAYIO_SMODE) WSDISPLAYIO_MODE_EMUL");
  }
}

static struct _platinfo *platmatch(char *name, int depth)
{
  struct _platinfo *p;

  p = &platinfo[0];
  while (p->depth) {
    if (p->depth == depth 
	&& strncmp(name, p->name, strlen(p->name)) == 0) {
      return p;
    }
    p++;
  }
  return NULL;
}

static int fbsetinfo(struct wsdisplay_fbinfo *fbinfo)
{
  int fbsize = 0;
  struct _platinfo *p;
  int mib[2];
  char *cp,*ep;
  size_t len;

  platform = 0;
  if ((ep = getenv("PLATFORM")) == NULL) {
    mib[0] = CTL_HW;
    mib[1] = HW_MODEL;
    sysctl(mib, 2, NULL, &len, NULL, 0);
    if ((cp = malloc(len)) != NULL) {
      sysctl(mib, 2, cp, &len, NULL, 0);
      while (*cp && *cp != ' ')       /* skip vender name */
	cp++;
      if (*cp) {
	cp++;
	platform = cp;
      }
    }
  } else {
    platform = ep;
  }
  printf("md_hpcmips fbsetinfo: platform = %s \n", platform);

  depth = fbinfo->depth;
  switch (depth) {
  case 8:
    rowbytes = DEF_ROWBYTE8;
    break;
  default:
    text_mode();
    printf("hpcmips fbsetinfo: unknown depth %d\n", fbinfo->depth);                
    return -1;
    break;
  }

  if (platform == NULL) {
    printf("hpcmips fbsetinfo: PLATFORM is not set; use default\n");
  } else if ((p = platmatch(platform, fbinfo->depth)) == NULL) {
    printf("hpcmips fbsetinfo: unknown platform %s\n", platform);                   
    return -1;
  } else {
    rowbytes = p->rowbyte;
    fb_off = p->off;
  }
  printf("display: %dx%dx%d\n",
	 fbinfo->width,
	 fbinfo->height,
	 fbinfo->depth);

  fbsize = rowbytes * fbinfo->height;

  printf("fbsize: %d off:%d rowbytes:%d\n", fbsize, fb_off, rowbytes);

  if (fb_off)
    fbsize += fb_off;
  return fbsize;
}


int open_hpcfb(void)
{
  int fbsize, dispmode;
  struct wsdisplay_fbinfo fbinfo;
  struct wsdisplay_cmap cmap;

  if((fb_fd = open(FBDEV,O_RDWR)) < 0) {
    perror(FBDEV);
    return(-1);
  }

  if(ioctl(fb_fd, WSDISPLAYIO_GINFO, &fbinfo) < 0) {
    perror("ioctl(WSDISPLAYIO_GINFO)");
    return(-1);
  }

  fbsize = fbsetinfo(&fbinfo);
  width = fbinfo.width;
  height = fbinfo.height;

  dispmode = WSDISPLAYIO_MODE_MAPPED;
  if(ioctl(fb_fd, WSDISPLAYIO_SMODE, &dispmode) < 0) {
    perror("ioctl(WSDISPLAYIO_SMODE) WSDISPLAYIO_MODE_MAPPED");
    return(-1);
  }

  fb = mmap(NULL, fbsize, PROT_READ | PROT_WRITE, MAP_SHARED, fb_fd, 0);
  if(fb == MAP_FAILED) {
    text_mode();
    perror("mmap()");
    return(-1);
  }
  if (fb_off)
    fb += fb_off;

#ifndef MG2_DEPTH16
  cmap.index = 0;
  cmap.count = 256;
  cmap.red = red;
  cmap.green = green;
  cmap.blue = blue;
  if(ioctl(fb_fd, WSDISPLAYIO_GETCMAP, &cmap) < 0) {
    text_mode();
    perror("GETCMAP");
    return(-1);
  }
#endif
  
  return(0);
}


void close_hpcfb(void)
{
  text_mode();
  return;
}


static inline int power2(int v)
{
  return v * v;
}


int get_color(u_char r, u_char g, u_char b)
{
#ifdef MG2_DEPTH16

  int red, green, blue;
  
  red = r;
  green = g;
  blue = b;

  #ifdef ORD_R5G5B5

    if(red)
      red = red * 2 + 1;
    if(green)
      green = green * 2 + 1;
    if(blue)
      blue = blue * 2 + 1;

    return ((red << 10) | (green << 5) | blue);

  #else

    if(red)
      red = red * 2 + 1;
    if(green)
      green = 0x3f + (green + 1) / 16;
    if(blue)
      blue = blue * 2 + 1;

    return ((red << 11) | (green << 5) | blue);

  #endif /* ORD_R5G5B5 */

#else

  int i, index;
  int delta, odelta;

  odelta = 65536;
  index = 0;
  for(i = 0; i < 256; i++) {
    delta = (power2(red[i] - r))   +
	    (power2(green[i] - g)) +
	    (power2(blue[i] - b))  / 3;

    if(delta < odelta) {
      odelta = delta;
      index = i;
      if(delta == 0)
        break;
    }
  }

  return index;

#endif /* MG2_DEPTH16 */
}

void clear_fmsx_screen(int dummy)
{
  pixel *line_buf;
  char *dstp;
  int i, col;

#ifdef MG2_DEPTH16
  line_buf = malloc(width * sizeof(short));
#else
  line_buf = malloc(width);
#endif

  if(!line_buf)
    return;

  col = get_color(0, 0, 0);
  for(i = 0; i < width; i++) {
    line_buf[i] = col;
  }

  dstp = fb;
  for(i = 0; i < height; i++) {

#ifdef MG2_DEPTH16
    memcpy(dstp, line_buf, width * sizeof(short));
#else
    memcpy(dstp, line_buf, width);
#endif

    dstp += rowbytes;
  }

  free(line_buf);
}


void bitblt(char *src, int w,int h)
{
  int y;
  char *srcp, *dstp;
  int w16b = 0;

  srcp = src;
  dstp = fb;

#ifdef MG2_DEPTH16
  w16b = w * sizeof(short);

  for(y = 0; y < h; y++) {
    memcpy(dstp, srcp, w16b);

#else

  for(y = 0; y < h; y++) {
    memcpy(dstp, srcp, w);

#endif

    dstp += rowbytes;
    srcp += w;
  }
}

void init_tty(void)
{
  tcgetattr(0, &save_tty);
  save_tty.c_iflag |= INLCR | ICRNL;
  save_tty.c_lflag |= ICANON | ECHO;
  save_tty.c_oflag |= OPOST | ONLCR;
  cur_tty = save_tty;

  cur_tty.c_iflag = IGNPAR | IGNBRK;
  cur_tty.c_oflag = ONLCR;
  cur_tty.c_cflag = CREAD | CS8;
  cur_tty.c_lflag &= ~(ICANON | IEXTEN | ECHO | ISIG);
  cur_tty.c_cc[VTIME] = 0;
  cur_tty.c_cc[VMIN] = 1;
  tcsetattr(0, TCSANOW, &cur_tty);
}

void reset_tty(void)
{
  tcsetattr(0, TCSANOW, &save_tty);
}



#define SET_KBMODE(fd, mode)              \
  {                                       \
    if(ioctl(fd, KDSKBMODE, mode) < 0)    \
    {                                     \
      perror("ioctl");                    \
      return -1;                          \
    }                                     \
    return 0;                             \
  }


int kbmode_raw(int fd)
{
  SET_KBMODE(fd, K_RAW);
}


int kbmode_normal(int fd)
{
  SET_KBMODE(fd, K_XLATE);
}


int get_key(void)
{
  fd_set fds;
  struct timeval to;
  int r;
  unsigned char dt;

  FD_ZERO(&fds);
  FD_SET(0,&fds);
  to.tv_sec = 0;
  to.tv_usec = 0;
  dt = 0;
  if (select(1,&fds,0,0,&to) != 1) return 0;
  r = read(0,&dt,sizeof(unsigned char));
  return (int)dt;
}




/** InitLibUnixMG2()  ****************************************/
/** This function is called to obtain the display and other **/
/** values at startup. It returns 0 on failure.             **/
/*************************************************************/
int InitLibUnixMG2(void)
{
  int J;

  /* Timers */  
  FirstTimer=NULL;
  TimerLock=0;

  return(1);
}

/** MasterTimer() ********************************************/
/** The main timer handler which is called MAXTIMERFREQ     **/
/** times a second. It then calls user-defined timers.      **/
/*************************************************************/
static void MasterTimer(int Arg)
{
  static unsigned int Counter=0;
  register struct Timer *P;

  if(!TimerLock)
    for(P=FirstTimer;P;P=P->Next)
      if(!(Counter%P->Freq)) (P->Handler)();

  Counter++;
  signal(Arg,MasterTimer);
}


/** TimerSignal() ********************************************/
/** Establish a signal handler called with given frequency  **/
/** (Hz). Returns 0 if failed.                              **/
/*************************************************************/
int TimerSignal(int Freq,void (*Handler)(int))
{
  struct itimerval TimerValue;

  TimerValue.it_interval.tv_sec  =
  TimerValue.it_value.tv_sec     = 0;
  TimerValue.it_interval.tv_usec =
  TimerValue.it_value.tv_usec    = 1000000L/Freq;
  if(setitimer(ITIMER_REAL,&TimerValue,NULL)) return(0);
  signal(SIGALRM,Handler);
  return(1);
}

/** AddTimer() ***********************************************/
/** Establish a periodically called routine at a given      **/
/** frequency (1..MAXTIMERFREQ Hz). Returns 0 if failed.    **/
/*************************************************************/
int AddTimer(int Freq,void (*Handler)(void))
{
  struct Timer *P;

  /* Freq must be 1..MAXTIMERFREQ */
  if((Freq<1)||(Freq>MAXTIMERFREQ)) return(0);

  /* Lock the timer queue */
  TimerLock=1;

  /* Look if this timer is already installed */
  for(P=FirstTimer;P;P=P->Next)
    if(P->Handler==Handler)
    { P->Freq=Freq;TimerLock=0;return(1); }

  /* Make up a new one if not */
  if(FirstTimer)
  {
    for(P=FirstTimer;P->Next;P=P->Next);
    P->Next=malloc(sizeof(struct Timer));
    P=P->Next;
  }
  else
  {
    /* Allocate the first timer */
    FirstTimer=malloc(sizeof(struct Timer));
    P=FirstTimer;

    /* Start up the main handler */
    if(P)
      if(!TimerSignal(MAXTIMERFREQ,MasterTimer))
      { free(P);P=NULL; }
  }

  /* Set up the timer and exit */
  if(P) { P->Next=0;P->Handler=Handler;P->Freq=Freq; }
  TimerLock=0;
  return(P!=0);
}

/** DelTimer() ***********************************************/
/** Remove routine added with AddTimer().                   **/
/*************************************************************/
void DelTimer(void (*Handler)(void))
{
  struct itimerval TimerValue;
  struct Timer *P,*T;

  TimerLock=1;

  if(FirstTimer)
    if(FirstTimer->Handler==Handler)
    {
      /* Delete first timer and free the space */
      P=FirstTimer;FirstTimer=P->Next;free(P);

      /* If no timers left, stop the main handler */
      if(!FirstTimer)
      {
         TimerValue.it_interval.tv_sec  =
         TimerValue.it_value.tv_sec     =
         TimerValue.it_interval.tv_usec =
         TimerValue.it_value.tv_usec    = 0;
         setitimer(ITIMER_REAL,&TimerValue,NULL);
         signal(SIGALRM,SIG_DFL);
      }
    }
  else
  {
    /* Look for our timer */
    for(P=FirstTimer;P;P=P->Next)
      if(P->Next->Handler==Handler) break;

    /* Delete the timer and free the space */
    if(P) { T=P->Next;P->Next=T->Next;free(T); }
  }

  TimerLock=0;
}

#endif /* UNIX */
