/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           UnixMG2.c                     **/
/**                                                         **/
/** This file contains NetBSD/hpcmips-dependent subroutines **/
/** and drivers. It includes common drivers from            **/
/** CommonMG2.h.                                            **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**               Hiroyuki Yanai  2000                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999, 2000
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 *      Shin Takemura (takemura@netbsd.org)
 *      SATO Kazumi (sato@netbsd.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifdef UNIX

/** Private #includes ****************************************/
#include "MSX.h"
#include "LibUnixMG2.h"

/** Standard Unix #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

/** Public parameters ************************************************/
char *Title="fMSX NetBSD/MG2 2.1";  /* Window/command line title     */
int SaveCPU  = 1;                   /* 1 = freeze when focus out     */
int UseSHM   = 1;                   /* 1 = use MITSHM                */
int UseZoom  = 1;                   /* Window zoom factor            */
int SyncFreq = 0;                   /* Screen update frequency       */
char *Disks[2][MAXDISK+1];          /* Disk names for each drive     */
int UseVDPWait  = 1;                /* 1 = use VDPWait               */
int RenshaWait  = 0;                /* Rensha Wait Count (n * vsync) */


/** Various variables *********************************************/
#define WIDTH  324
#define HEIGHT 228

static volatile byte TimerReady;
static unsigned int BPal[256],XPal[16],XPal0; 
static byte JoyState;
static int CurDisk[2];

/** Various MG2-related definitions **************************/
#define WIDTH512 512
static char *offscrn;
static int rensha_counter=0;


static inline int is_width512()
{
#ifndef NARROW
  static int retvals[MAXSCREEN+2]=
  {
    0, 0, 0, 0, 0, 0, 1, 1, 0, 1,
  };

  return retvals[ScrMode];
#else
  return 0;
#endif
}

#define MGK_F1       59
#define MGK_F2       60
#define MGK_F3       61
#define MGK_F4       62
#define MGK_F5       63
#define MGK_F6       64
#define MGK_F7       65
#define MGK_F8       66
#define MGK_F9       67
#define MGK_F10      68

#define MGK_UP       72
#define MGK_DOWN     80
#define MGK_LEFT     75
#define MGK_RIGHT    77

#define MGK_PAGEUP   73
#define MGK_PAGEDOWN 81

#define MGK_ALT      56
#define MGK_ZENHAN   41
#define MGK_HIRAKATA 112
#define MGK_XFER     121
#define MGK_NFER     123
#define MGK_DEL      14
#define MGK_CAPS     58
#define MGK_SHIFT    42
#define MGK_TAB      15
#define MGK_ENTER    28
#define MGK_BS       14
#define MGK_CTRL     29

static int strig1, strig2;


/*
    hpcmips  key map ˴ؤ륳
    
    fMSX ü쥭ȤбϰʲΤȤꡣ

        F6                -> STOP
        F7                -> CLS
        CTRL+F7           -> change Disk-Image (drive a)
        CTRL+F8           -> change Disk-Image (drive b)
        F9                -> built-in debugger
        F10               -> quit fMSX
        PageUp            -> INS
        PageDown          -> DEL
        Zen/Hnakaku       -> SELECT
        Hiragana/Katakana -> kana
        XFER              -> GRAPH
        CAPS              -> CAPS

        ALT               -> TRIGGER1
        NFER              -> TRIGGER2
*/
static byte mg2_keymap[256] =
{
/* 00  */  0x00, 0x1B, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, /* - ESC 1 2 3 4 5 6           */
/* 08  */  0x37, 0x38, 0x39, 0x30, 0x2d, 0x3d, 0x08, 0x09, /* 7 8 9 0 - ^ BS TAB          */
/* 16  */  0x71, 0x77, 0x65, 0x72, 0x74, 0x79, 0x75, 0x69, /* q w e r t y u i             */

#ifndef SWAP_CTRLCAPS
/* 24  */  0x6f, 0x70, 0x5b, 0x5d, 0x0d, 0xe3, 0x61, 0x73, /* o p @ [ RET CTRL a s        */
#else
/* 24  */  0x6f, 0x70, 0x5b, 0x5d, 0x0d, 0xe2, 0x61, 0x73, /* o p @ [ RET CAPS a s        */
#endif

/* 32  */  0x64, 0x66, 0x67, 0x68, 0x6a, 0x6b, 0x6c, 0x3b, /* d f g h j k l ;             */
/* 40  */  0x27, 0xc6, 0xe1, 0x60, 0x7a, 0x78, 0x63, 0x76, /* : SELECT SHIFT ] z x c v    */
/* 48  */  0x62, 0x6e, 0x6d, 0x2c, 0x2e, 0x2f, 0x00, 0x00, /* b n m , . / - -             */

#ifndef SWAP_CTRLCAPS
/* 56  */  0x00, 0x20, 0xe2, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, /* - SPACE CAPS F1 F2 F3 F4 F5 */
#else
/* 56  */  0x00, 0x20, 0xe3, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, /* - SPACE CTRL F1 F2 F3 F4 F5 */
#endif

/* 64  */  0xc7, 0xc3, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* STOP CLS - - - - - -        */
/* 72  */  0x52, 0xc5, 0x00, 0x51, 0x00, 0x53, 0x00, 0x00, /* UP INS - LEFT - RIGHT - -   */
/* 80  */  0x54, 0xc4, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* DOWN DEL - - - - - -        */
/* 88  */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 96  */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 104 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 112 */  0xe4, 0x00, 0x00, 0xf0, 0x00, 0x00, 0x00, 0x00, /* KANA - - _ - - - -          */
/* 120 */  0x00, 0xea, 0x00, 0x00, 0x00, 0x5c, 0x00, 0x00, /* - GRAPH - - - \ - -         */
/* 128 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 136 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 144 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 152 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 160 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 168 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 176 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 184 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 192 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 200 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 208 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 216 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 224 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 232 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 240 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
/* 248 */  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* - - - - - - - -             */
};


/** Sound-related definitions ********************************/
#ifdef SOUND
int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void StopSound(void);
void ResumeSound(void);
extern int UseSound;
#endif

/** These functions are called on signals ********************/
static void OnBreak(int Arg) { ExitNow=1;signal(Arg,OnBreak); }
static void OnTimer(int Arg) { TimerReady=1;signal(Arg,OnTimer); }

/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{6,0x10},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {2,0x20},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};



/** InitMachine() ********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  int J;

  /* Reset all variables */
  offscrn=NULL;JoyState=0x00;
  CurDisk[0]=CurDisk[1]=0;

  /* Correct configuration values */
  UseZoom  = UseZoom>4? 4:UseZoom<1? 1:UseZoom;
  SyncFreq = SyncFreq>60? 60:SyncFreq<0? 0:SyncFreq;

  /* Open display */
  if(Verbose) printf("Initializing NetBSD/hpcmips drivers:\n  Opening display...");
  if(open_hpcfb()) { if(Verbose) puts("FAILED");return(0); }

  /* Reset the palette */
  XPal0=0;
  for(J=0;J<16;J++) XPal[J]=0;

  /* Initialize LibUnix toolkit */
  if(Verbose) printf("OK\n  Initializing LibUnixMG2...");
  if(!InitLibUnixMG2()) { if(Verbose) puts("FAILED");return(0); }

  /* Initialize image buf */
  if(Verbose) printf("OK\n  Initializing MG2 offscreen...");

#ifdef NARROW
  offscrn = malloc(WIDTH * HEIGHT);
#else
  offscrn = malloc(WIDTH512 * HEIGHT);
#endif

  if(!offscrn) { if(Verbose) puts("FAILED");return(0); }
  if(Verbose) printf("OK\n");

  /* clear screen */
  clear_fmsx_screen(0);

  /* init ttys */
  kbmode_raw(0);
  init_tty();

  /* open wsmouse */
  open_mouse();

  /* Allocate colors for SCREEN8 */
  {
    int R, G, B;
    int I;

    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          I=B|(R<<3)|(G<<6);

          BPal[I]      =
          BPal[I|0x04] =
          BPal[I|0x20] =
          BPal[I|0x24] =
	    get_color((R * 64)& 0xff, (G * 64) & 0xff, (B * 64) & 0xff);
        }
  }

#ifdef SOUND
  /* Initialize sound */
  if(UseSound) InitSound();
#endif SOUND

  /* Catch all signals */
  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  /* Establish timer signal if needed */
  if(SyncFreq) TimerSignal(SyncFreq,OnTimer);

  return(1);
}

/** TrashMachine() *******************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  unsigned long C;
  int J;

  if(Verbose) printf("Shutting down...\n");

#ifdef SOUND
  StopSound();TrashSound();
#endif SOUND

  /* dispose offscrn */
  if(offscrn) {
      free(offscrn);
  }
  
  /* reset ttys */
  kbmode_normal(0);
  reset_tty();

  /* close wsmouse */
  close_mouse();

  close_hpcfb();
}

/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{
  /* Wait if needed */
  if(SyncFreq) while(!TimerReady&&!ExitNow);

  /* Copy image */
  if(is_width512()) {
    bitblt(offscrn, WIDTH512, HEIGHT);
  } else {
    bitblt(offscrn, WIDTH, HEIGHT);
  }

  /* Reset timer flag */
  TimerReady=0;
}

/** Keyboard() ***********************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
  static byte Control=0;
  static byte TriggerState=0;
  int J,I,c;

#ifdef SOUND
  /* FlushSound(); */  /* Flush sound stream on each interrupt */
#endif

  /* Check for keypresses/keyreleases */
  c=get_key();
  J=c&0x7f;
  if(!(c&0x80))
  {
    /* Special keys pressed... */
    switch(J)
    {
#ifdef DEBUG
      /* case MGK_F9:   CPU.Trace=!CPU.Trace;break; */
#endif

      case MGK_F7:
      case MGK_F8:
        if(Control)
        {
          I=(J==MGK_F8)? 1:0;
          if(Disks[I][0])
          {
            CurDisk[I]=(CurDisk[I]+1)%MAXDISK;
            if(!Disks[I][CurDisk[I]]) CurDisk[I]=0;
            ChangeDisk(I,Disks[I][CurDisk[I]]);
          }
        }
        break;

      case MGK_F10:   ExitNow=1;break;
      case MGK_CTRL:  Control=1;break;

      /*  joy stick */
      case MGK_UP:    JoyState|=0x01;break;
      case MGK_DOWN:  JoyState|=0x02;break;
      case MGK_LEFT:  JoyState|=0x04;break;
      case MGK_RIGHT: JoyState|=0x08;break;
      case MGK_ALT:   JoyState|=0x10;strig1=1;break;
      case MGK_NFER:  JoyState|=0x20;strig2=1;break;
    }

    /* Key pressed: reset bit in KeyMap */
    J=mg2_keymap[J];
    J&=0xFF;KeyMap[Keys[J][0]]&=~Keys[J][1];
  }
  else
  {
    /* Special keys released... */
    switch(J)
    {
      case MGK_CTRL:  Control=0;break;

      /* joy stick */
      case MGK_UP:    JoyState&=0xFE;break;
      case MGK_DOWN:  JoyState&=0xFD;break;
      case MGK_LEFT:  JoyState&=0xFB;break;
      case MGK_RIGHT: JoyState&=0xF7;break;
      case MGK_ALT:   JoyState&=0xEF;strig1=0;break;
      case MGK_NFER:  JoyState&=0xDF;strig2=0;break;
    }

    /* Key released: set bit in KeyMap */
    J=mg2_keymap[J];
    J&=0xFF;KeyMap[Keys[J][0]]|=Keys[J][1];
  }

  /* Rensha */
  if(RenshaWait && strig1)
  {
    rensha_counter++;
    if(rensha_counter>RenshaWait)
    {
      rensha_counter=0;
      TriggerState=(TriggerState&0x10)? (0):(0x10);
      JoyState=(JoyState&0xEF)|TriggerState;
    }
  }

}


/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/************************************ TO BE WRITTEN BY USER **/
byte Joystick(register byte N) { return(JoyState); }

/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/************************************ TO BE WRITTEN BY USER **/
int Mouse(register byte N)
{
  int X,Y,J,Buttons,widmax;

  widmax=WIDTH512-1;

  /* Get mouse coordinates */
  mouse_event();
  get_mouse_values(&X, &Y, &Buttons);

  if(!is_width512())
  {
    X-=(WIDTH-256)/2;
    widmax=255;
  }
  X=X<0? 0:(X>widmax? widmax:X);
  Y=Y<0? 0:(Y>211? 211:Y);

  Buttons=(strig1? 0x10000:0x00000) |
          (strig2? 0x20000:0x00000);

  return((Y<<8)|X|Buttons);
}

/** SetColor() ***********************************************/
/** Set color N (0..15) to R,G,B.                           **/
/*************************************************************/
void SetColor(register byte N,register byte R,register byte G,register byte B)
{
  XPal[0]=XPal0;

  /* Allocate a new color entry */
  XPal[N]=get_color(R, G, B);

  XPal0=XPal[0];
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
}

/** Dummy Functions ******************************************/
/** These dummy functions are called when writes to sound   **/
/** registers occur. Replace them with the actual functions **/
/** generating sound.                                       **/
/*************************************************************/
#ifndef SOUND
void PSGOut(register byte R,register byte V)  { }
void SCCOut(register byte R,register byte V)  { }
#endif
void PPIOut(register byte Toggle)             { }
void OPLLOut(register byte R,register byte V) { }

/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "CommonMG2.h"

#endif /* UNIX */
