/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           UnixMGL2.c                    **/
/**                                                         **/
/** This file contains Unix/MGL2-dependent subroutines and  **/
/** drivers. It includes common drivers from CommonMGL2.h.  **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**               Elan Feingold   1995                      **/
/**               Ville Hallik    1996                      **/
/**               Hiroyuki Yanai  2000                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#ifdef UNIX

/** Private #includes ****************************************/
#include "MSX.h"
#include "LibUnixMGL2.h"

/** Standard Unix #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

/** Public parameters ********************************************/
char *Title="fMSX Unix/MGL2 2.1";  /* Window/command line title  */
int SaveCPU  = 1;               /* 1 = freeze when focus out     */
int UseSHM   = 1;               /* 1 = use MITSHM                */
int UseZoom  = 1;               /* Window zoom factor            */
int SyncFreq = 0;               /* Screen update frequency       */
char *Disks[2][MAXDISK+1];      /* Disk names for each drive     */
int UseVDPWait  = 1;            /* 1 = use VDPWait               */
int RenshaWait  = 2;            /* Rensha Wait Count (n * vsync) */

/** Various variables ****************************************/
#define WIDTH  324
#define HEIGHT 228

static volatile byte TimerReady;
static unsigned int BPal[256],XPal[16],XPal0; 
static byte JoyState;
static int CurDisk[2];

/** Various MGL2-related definitions *************************/
#define WIDTH512 512
static struct screen *offscr;
static struct virtual_key *vkey;

static inline int is_width512()
{
#ifndef NARROW
  static int retvals[MAXSCREEN+2]=
  {
    0, 0, 0, 0, 0, 0, 1, 1, 0, 1,
  };

  return retvals[ScrMode];
#else
  return 0;
#endif
}


struct mgl_key {
  int code;
  int shift;
  int control;
};

#define KQUEUE_SIZE 5
#define STRIG1 0x300
#define STRIG2 0x301
static struct mgl_key k_queue[KQUEUE_SIZE];
static int key_mode = 0;
static int strig1, strig2;

static void init_kqueue(void);


static struct mgl_key mgl_keymap[128] =
{
/* 00 */
  {-1  , 0, 0}, {0x61, 0, 1}, {0x62, 0, 1}, {0x63, 0, 1},  /*  C+a C+b C+c       */
  {0x64, 0, 1}, {0x65, 0, 1}, {0x66, 0, 1}, {0x67, 0, 1},  /*  C+d C+e C+f C+g   */
  {0x68, 0, 1}, {0x69, 0, 1}, {0x6a, 0, 1}, {0x6b, 0, 1},  /*  C+h C+i C+j C+k   */
  {0x6c, 0, 1}, {0x6d, 0, 1}, {0x6e, 0, 1}, {0x6f, 0, 1},  /*  C+l C+m C+n C+o   */

/* 10 */
  {0x70, 0, 1}, {0x71, 0, 1}, {0x72, 0, 1}, {0x73, 0, 1},  /*  C+p C+q C+r C+s   */
  {0x74, 0, 1}, {0x75, 0, 1}, {0x76, 0, 1}, {0x77, 0, 1},  /*  C+t C+u C+v C+w   */
  {0x78, 0, 1}, {0x79, 0, 1}, {0x7a, 0, 1}, {0x5d, 0, 1},  /*  C+x C+y C+z ESC   */
  {0x53, 0, 0}, {0x51, 0, 0}, {0x52, 0, 0}, {0x54, 0, 0},  /*  C+\ C+] C+^ C+_   */

/* 20 */
  {-1  , 0, 0}, {0x31, 1, 0}, {0x32, 1, 0}, {0x33, 1, 0},  /*  !   "   #          */
  {0x34, 1, 0}, {0x35, 1, 0}, {0x36, 1, 0}, {0x37, 1, 0},  /*  $   %   &   '      */
  {0x38, 1, 0}, {0x39, 1, 0}, {0x27, 1, 0}, {0x3b, 1, 0},  /*  (   )   *   +      */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},

/* 30 */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {0x27, 0, 0}, {-1  , 0, 0},  /*  :                   */
  {0x2c, 1, 0}, {0x2d, 1, 0}, {0x2e, 1, 0}, {0x2f, 1, 0},  /*  <   =   >   ?       */

/* 40 */
  {0x5b, 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},  /*  @                   */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},

/* 50 */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {0x5d, 0, 0},  /*  [                   */
  {-1  , 0, 0}, {0x60, 0, 0}, {0x3d, 0, 0}, {-1  , 0, 0},  /*  ]   ^               */

/* 60 */
  {0x5b, 1, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},  /*  `                   */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},

/* 70 */
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0},
  {-1  , 0, 0}, {-1  , 0, 0}, {-1  , 0, 0}, {0x5d, 1, 0},  /*  {                   */
  {0x5c, 1, 0}, {0x60, 1, 0}, {0x3d, 1, 0}, {0xff, 0, 0},  /*  |   }   ~   DEL     */
};

static struct mgl_key mgl_ctrl_keymap1[32] = {
/* 00 */
  {-1   , 0, 0}, {0xe3 , 0, 0}, {0x20 , 0, 0}, {0xc7 , 0, 1},  /* CTRL SPACE C+STOP    */
  {0xff , 0, 0}, {-1   , 0, 0}, {-1   , 0, 0}, {0xea , 0, 0},  /* DEL GRAPH            */
  {-1   , 0, 0}, {-1   , 0, 0}, {-1   , 0, 0}, {0xc3 , 0, 0},  /* HOME                 */
  {0xc3 , 1, 0}, {-1   , 0, 0}, {-1   , 0, 0}, {-1   , 0, 0},  /* CLS                  */

/* 10 */
  {-1   , 0, 0}, {0x300, 0, 0}, {0xc5 , 0, 0}, {-1   , 0, 0},  /* TRIG1 INS            */
  {0x301, 0, 0}, {-1   , 0, 0}, {0xc7 , 0, 0}, {-1   , 0, 0},  /* TRIG2 STOP           */
  {0xc6 , 0, 0}, {-1   , 0, 0}, {0xe1 , 0, 0}, {-1   , 0, 0},  /* SELECT SHIFT         */
  {-1   , 0, 0}, {-1   , 0, 0}, {-1   , 0, 0}, {-1   , 0, 0},  /*                      */
};  


#define INIT_KEYS(buf,size) {    \
    int i;                       \
    for(i=0;i<size;i++)          \
    {                            \
      buf[i].code = 0;           \
      buf[i].shift = 0;          \
      buf[i].control = 0;        \
    }                            \
}


/** Sound-related definitions ********************************/
#ifdef SOUND
int InitSound(void);
void TrashSound(void);
void FlushSound(void);
void StopSound(void);
void ResumeSound(void);
extern int UseSound;
#endif

/** These functions are called on signals ********************/
static void OnBreak(int Arg) { ExitNow=1;signal(Arg,OnBreak); }
static void OnTimer(int Arg) { TimerReady=1;signal(Arg,OnTimer); }

/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{6,0x10},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};


/** InitMachine() ********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  int J;

  /* Reset all variables */
  offscr=NULL;JoyState=0x00;
  CurDisk[0]=CurDisk[1]=0;

  /* Correct configuration values */
  UseZoom  = UseZoom>4? 4:UseZoom<1? 1:UseZoom;
  SyncFreq = SyncFreq>60? 60:SyncFreq<0? 0:SyncFreq;

  /* Open display */
  if(Verbose) printf("Initializing Unix/MGL2 drivers:\n  Opening display...");
  if(!open_graph()) { if(Verbose) puts("FAILED");return(0); }

  /* clear screen */
  clear_fmsx_screen(0);

  /* set mgl2's virtual_key */
  vkey = NULL;
  vkey = create_virtual_key3(0, 0, 639, 239, MK_V1, MK_V2, MK_V3);
  if(vkey)
  {
    vk_attach(NULL, vkey);
  }

  /* Reset the palette */
  XPal0=0;
  for(J=0;J<16;J++) XPal[J]=0;

  /* Initialize LibUnix toolkit */
  if(Verbose) printf("OK\n  Initializing LibUnixMGL2...");
  if(!InitLibUnixMGL2()) { if(Verbose) puts("FAILED");return(0); }

  /* Initialize image buf */
  if(Verbose) printf("OK\n  Initializing MGL2 offscreen...");

#ifdef NARROW
  offscr=create_memscreen(WIDTH, HEIGHT, NULL, STK_NATIVE, 0);
#else
  offscr=create_memscreen(WIDTH512, HEIGHT, NULL, STK_NATIVE, 0);
#endif

  if(!offscr) { if(Verbose) puts("FAILED");return(0); }
  if(Verbose) printf("OK\n");
  push_screen(offscr);

  /* Allocate colors for SCREEN8 */
  {
    int R, G, B;
    int I;

    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          I=B|(R<<3)|(G<<6);

          BPal[I]      =
          BPal[I|0x04] =
          BPal[I|0x20] =
          BPal[I|0x24] =
            mc_from_rgb(packRGB(R*4, G*4, B*4));
        }  
  }

#ifdef SOUND
  /* Initialize sound */
  if(UseSound) InitSound();
#endif SOUND

  /* Catch all signals */
  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  /* Establish timer signal if needed */
  if(SyncFreq) TimerSignal(SyncFreq,OnTimer);

  
  /* init key queue */
  init_kqueue();

  return(1);
}

/** TrashMachine() *******************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  unsigned long C;
  int J;

  if(Verbose) printf("Shutting down...\n");

#ifdef SOUND
  StopSound();TrashSound();
#endif SOUND

  /* dispose offscr */
  pop_screen();
  if(offscr) {
      free_screen(offscr);
  }
  close_graph();
}

/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{
  /* Wait if needed */
  if(SyncFreq) while(!TimerReady&&!ExitNow);
  /* Copy image */
  pop_screen();

  if(is_width512())
  {
    bitblt(NULL, 0, 0, offscr, 0, 0, WIDTH512, HEIGHT, 0);
  }
  else
  {
    bitblt(NULL, 0, 0, offscr, 0, 0, WIDTH, HEIGHT, 0);
  }

  push_screen(offscr);
  /* Reset timer flag */
  TimerReady=0;
}

/** Keyboard() ***********************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
static void init_kqueue(void)
{
  strig1 = strig2 = 0;
  INIT_KEYS(k_queue,KQUEUE_SIZE)
}


static int put_kqueue(struct mgl_key k)
{
  /* $B%-%e!<$,0n$l$F$$$k$J$i!"%-%e!<$KF~$l$:!"(B0 $B$rJV$9(B */
  if(k_queue[KQUEUE_SIZE - 1].code != 0)
  {
    return 0;
  }
  k_queue[KQUEUE_SIZE - 1] = k;
  return 1;
}


static void get_kqueue(struct mgl_key *k)
{
  *k = k_queue[0];

  memmove(k_queue, k_queue + 1, (KQUEUE_SIZE - 1) * sizeof(struct mgl_key));
  k_queue[KQUEUE_SIZE - 1].code = 0;
  k_queue[KQUEUE_SIZE - 1].shift = 0;
  k_queue[KQUEUE_SIZE - 1].control = 0;
}


static void flush_kqueue()
{
  int i;

  init_kqueue();

  /* MSX $B$J%-!<%^%H%j%/%9$r=i4|2=(B */
  for(i=0;i<16;i++)  KeyMap[i]=0xff;
}


static void bind_key(int *key, int *shift, int *control)
{
  int bkey;
  int I;
  struct mgl_key *keymap;

  bkey = -1;
  *shift = *control = 0;
  if(*key < 0x80)
  {
    keymap = ((key_mode == 1) && (*key < 0x20)) ? (mgl_ctrl_keymap1) : (mgl_keymap);
    bkey = (*(keymap + *key)).code;
    *shift = (*(keymap + *key)).shift;
    *control = (*(keymap + *key)).control;

    /*  check & conv upper case alphabet chars */
    if(isalpha(*key) && isupper(*key))
    {
      bkey = *key + 0x20;
      *shift = 1;
    }
  }

  if(bkey != -1)  
  {
    *key = bkey;
  }

  if(*key<0x100)  return;

  /* Special keys pressed... */
  switch(*key)
  {
#ifdef DEBUG
/*  case MK_F10:   CPU.Trace=!CPU.Trace;break; */
#endif
    case MK_F1:
    case MK_F2:
    case MK_F3:
    case MK_F4:
    case MK_F5:
      *key = (*key - MK_F1) + 0xBE;                 /* F1 .. F5 */
      break;

    case MK_F10:       ExitNow=1;break;

    case MK_HOME:      *key=0xC3;break;             /* home     */

    case MK_DEL:       *key=0xC4;break;             /* del      */
    case MK_INS:       *key=0xC5;break;             /* ins      */
    case MK_PAGE_UP:   *key=0xE4;break;             /* kana     */
    case MK_PAGE_DOWN: *key=0xC4;break;             /* del      */

    case MK_UP:        *key=0x52;break;
    case MK_DOWN:      *key=0x54;break;
    case MK_LEFT:      *key=0x51;break;
    case MK_RIGHT:     *key=0x53;break;

    /* mouse buttons */
//    case MK_V1:                                     /* btn1 pressed (fall through) */
//    case MK_V2:        *key=0;strig1=-1;break;      /* btn1 down */
//    case MK_V3:        *key=0;strig1=0; break;      /* btn1 up   */

    case STRIG1:       *key=0;strig1=~strig1;break; /* btn1 pressed (toggle) */
    case STRIG2:       *key=0;strig2=~strig2;break; /* btn2 pressed (toggle) */

    default:           *key=0;break;
  }
}


static void set_keymap(int key, int shift, int control)
{
  /* Key released: set bit in KeyMap */
  if(shift)
  {
    KeyMap[6]|=0x01;	/* ---- 6,0x01 .... SHIFT */
  }
  if(control)
  {
    KeyMap[6]|=0x02;	/* ---- 6,0x02 .... CONTROL */
  }
  key&=0xFF;KeyMap[Keys[key][0]]|=Keys[key][1];
}


static void reset_keymap(int key, int shift, int control)
{
  /* Key pressed: reset bit in KeyMap */
  if(shift) 
  {
    KeyMap[6]&=~0x01;	/* ---- 6,0x01 .... SHIFT */
  }
  if(control)
  {
    KeyMap[6]&=~0x02;	/* ---- 6,0x02 .... CONTROL */
  }
  key&=0xFF;KeyMap[Keys[key][0]]&=~Keys[key][1];
}

void Keyboard(void)
{
  int J,I;
  int bkey, shift, control;
  struct mgl_key key, key2;
  static int isLockMode = 0;

#ifdef SOUND
  FlushSound();  /* Flush sound stream on each interrupt */
#endif

  key.shift = 0;
  key.control = 0;
  key.code = 0;

  /* Check for keypresses */
  if((J = get_key(0)) >= 0)
  {
    key.code = J;

    if(J>0xff)
    {
      /* $B%-!<%^%C%T%s%0$NJQ99(B */
      if(J==MK_F6)
      {
        key_mode++;
        if(key_mode > 1)  key_mode = 0;
        flush_kqueue();
        key.code=J=0;
      }

      /* $B%-!<%m%C%/%b!<%I(B */
      if(J==MK_F7)
      {
        isLockMode=~isLockMode;
        if(isLockMode == 0)
        {
          for(I=0;I<16;I++)
          {
            KeyMap[I]=0xff;
          }
        }
        flush_kqueue();
        key.code=J=0;
      }

      /* $B%G%#%9%/%$%a!<%8$N@ZBX(B */
      if(J==MK_F8 || J==MK_F9)
      {
        I=(J==MK_F9)? 1:0;
        if(Disks[I][0])
         {
           CurDisk[I]=(CurDisk[I]+1)%MAXDISK;
           if(!Disks[I][CurDisk[I]]) CurDisk[I]=0;
           ChangeDisk(I,Disks[I][CurDisk[I]]);
         }
         key.code=J=0;
      }
    }

    /*
      $B%-%e!<$KF~$l$k;v$,$G$-$?$J$i!"%-!<%^%H%j%/%9$r%j%;%C%H!J3:Ev%-!<$,2!$5$l$k!K$9$k!#(B
      $B$=$&$G$J$1$l$P!"F~NO$rL5;k$9$k!J<N$F$k!K!#(B
    */
    if(put_kqueue(key))
    {
      bind_key(&J, &shift, &control);
      reset_keymap(J, shift, control);
    }
  }
  else 
  {
    /* key released... */
    get_kqueue(&key);
    if(isLockMode == 0 || key.code > 0x1f)
    {
      bkey = key.code;
      bind_key(&bkey, &shift, &control);
      set_keymap(bkey, shift, control);
    }
  }

#ifdef SOUND
  /* StopSound(); */
#endif
  
#ifdef SOUND
  /* ResumeSound(); */
#endif
}


/** Joystick() ***********************************************/
/** Query position of a joystick connected to port N.       **/
/** Returns 0.0.F2.F1.R.L.D.U.                              **/
/************************************ TO BE WRITTEN BY USER **/
byte Joystick(register byte N) { return(JoyState); }

/** Mouse() **************************************************/
/** Query coordinates of a mouse connected to port N.       **/
/** Returns F2.F1.Y.Y.Y.Y.Y.Y.Y.Y.X.X.X.X.X.X.X.X.          **/
/************************************ TO BE WRITTEN BY USER **/
int Mouse(register byte N)
{
  int X,Y,J,Buttons,widmax;

  Buttons=0;

  /* Get mouse coordinates */
  if(!is_width512())
  {
    widmax=255;
    X=vk_x-((WIDTH-256)/2);
  }
  else
  {
    widmax=WIDTH512-1;
    X=vk_x;
  }
  Y=vk_y;
  X=X<0? 0:(X>widmax? widmax:X);
  Y=Y<0? 0:(Y>211? 211:Y);

  Buttons=(strig1? 0x10000:0x00000) |
          (strig2? 0x20000:0x00000);

  return (Y<<8)|X|Buttons;
}

/** SetColor() ***********************************************/
/** Set color N (0..15) to R,G,B.                           **/
/*************************************************************/
void SetColor(register byte N,register byte R,register byte G,register byte B)
{
  XPal[0]=XPal0;

  /* Allocate a new color entry */
  R>>=4; G>>=4; B>>=4;
  XPal[N]=mc_from_rgb(packRGB(R, G, B));

  XPal0=XPal[0];
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
}

/** Dummy Functions ******************************************/
/** These dummy functions are called when writes to sound   **/
/** registers occur. Replace them with the actual functions **/
/** generating sound.                                       **/
/*************************************************************/
#ifndef SOUND
void PSGOut(register byte R,register byte V)  { }
void SCCOut(register byte R,register byte V)  { }
#endif
void PPIOut(register byte Toggle)             { }
void OPLLOut(register byte R,register byte V) { }

/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "CommonMGL2.h"

#endif /* UNIX */
