/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                          fMSX.c                         **/
/**                                                         **/
/** This file contains generic main() procedure statrting   **/
/** the emulation.                                          **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "MSX.h"
#include "Help.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

char *Options[]=
{ 
  "verbose","hperiod","vperiod","uperiod","help","printer",
  "serial","diska","diskb","tape","font","logsnd","ram","vram",
  "rom","msx1","msx2","msx2+","joy","trap","sound","nosound",
  "shm","noshm","saver","nosaver","sync","nosync","scale",
  "vdpwait","novdpwait","afire","noafire",0
};

extern char *Title;      /* Program title                       */
extern int   UseSound;   /* Sound mode (#ifdef SOUND)           */
extern int   UseSHM;     /* Use SHM X extension (#ifdef MITSHM) */
extern int   UseZoom;    /* Zoom factor (#ifdef UNIX)           */
extern int   SaveCPU;    /* Pause when inactive (#ifdef UNIX)   */
extern int   SyncScreen; /* Sync screen updates (#ifdef MSDOS)  */
extern int   SyncFreq;   /* Sync screen updates (#ifdef UNIX)   */

/** Zero-terminated arrays of disk names for each drive ******/
extern char *Disks[2][MAXDISK+1];

/* Use VDP Wait */
extern int   UseVDPWait;

/* Use Rensha */
extern int   RenshaWait; /* Rensha Wait Count (n * vsync) */


/** main() ***************************************************/
/** This is a main() function used in Unix and MSDOS ports. **/
/** It parses command line arguments, sets emulation        **/
/** parameters, and passes control to the emulation itself. **/
/*************************************************************/
int main(int argc,char *argv[])
{
  int CartCount,TypeCount;
  int JoyCount,DiskCount[2];
  int N,J;

#ifdef DEBUG
  CPU.Trap=0xFFFF;CPU.Trace=0;
#endif

#ifdef MSDOS
  Verbose=0;
#else
  Verbose=1;
#endif

  /* Clear everything */
  CartCount=TypeCount=JoyCount=0;
  DiskCount[0]=DiskCount[1]=0;

  /* Default disk images */
  Disks[0][1]=Disks[1][1]=0;
  Disks[0][0]=DiskA;
  Disks[1][0]=DiskB;

  for(N=1;N<argc;N++)
    if(*argv[N]!='-')
      switch(CartCount++)
      {
        case 0: CartA=argv[N];break;
        case 1: CartB=argv[N];break;
        default: printf("%s: Excessive filename '%s'\n",argv[0],argv[N]);
      }
    else
    {    
      for(J=0;Options[J];J++)
        if(!strcmp(argv[N]+1,Options[J])) break;

      switch(J)
      {
        case 0:  N++;
                 if(N<argc) Verbose=atoi(argv[N]);
                 else printf("%s: No verbose level supplied\n",argv[0]);
                 break;
        case 1:  N++;
                 if(N>=argc)
                   printf("%s: No HBlank period supplied\n",argv[0]);
                 else
                 {
                   J=atoi(argv[N]);
                   if((J>=30)&&(J<=3000)) HPeriod=J;
                 }
                 break;
        case 2:  N++;
                 if(N>=argc)
                   printf("%s: No VBlank period supplied\n",argv[0]);
                 else
                 {
                   J=atoi(argv[N]);
                   if((J>=1000)&&(J<=100000)) VPeriod=J;
                   if((J>=256)&&(J<=999)) VPeriod=HPeriod*J;
                 }
                 break;
        case 3:  N++;
                 if(N>=argc)
                   printf("%s: No screen update period supplied\n",argv[0]);
                 else
                 {
                   J=atoi(argv[N]);
                   if((J>=1)&&(J<=20)) UPeriod=J; 
                 }
                 break;
	case 4:  printf
                 ("%s by Marat Fayzullin    (C)FMS 1994-1999\n",Title);
                 for(J=0;HelpText[J];J++) puts(HelpText[J]);
                 return(0);
        case 5:  N++;
                 if(N<argc) PrnName=argv[N];
                 else printf("%s: No file for printer redirection\n",argv[0]);
                 break;
        case 6:  N++;
                 if(N<argc) ComName=argv[N];
                 else printf("%s: No file for serial redirection\n",argv[0]);
                 break;
        case 7:  N++;
                 if(N>=argc)
                   printf("%s: No file for drive A\n",argv[0]);
                 else
                   if(DiskCount[0]>=MAXDISK)
                     printf("%s: Too many disks for drive A\n",argv[0]);
                   else
                     Disks[0][DiskCount[0]++]=argv[N];
                 break;
        case 8:  N++;
                 if(N>=argc) 
                   printf("%s: No file for drive B\n",argv[0]);
                 else
                   if(DiskCount[1]>=MAXDISK) 
                     printf("%s: Too many disks for drive B\n",argv[0]);
                   else
                     Disks[1][DiskCount[1]++]=argv[N];
                 break;
        case 9:  N++;  
                 if(N<argc) CasName=argv[N];
                 else printf("%s: No file for the tape\n",argv[0]);
                 break;
        case 10: N++;
                 if(N<argc) FontName=argv[N];  
                 else printf("%s: No font name supplied\n",argv[0]);
                 break;
        case 11: N++;
                 if(N<argc) SndName=argv[N];
                 else printf("%s: No file for soundtrack logging\n",argv[0]);
                 break;
        case 12: N++;
                 if(N>=argc)
                   printf("%s: No number of RAM pages supplied\n",argv[0]);
                 else
                   RAMPages=atoi(argv[N]);
                 break;
        case 13: N++;
                 if(N>=argc)
                   printf("%s: No number of VRAM pages supplied\n",argv[0]);
                 else
                   VRAMPages=atoi(argv[N]);
                 break;
        case 14: N++;  
                 if(N>=argc)
                   printf("%s: No ROM mapper type supplied\n",argv[0]);
                 else
                   switch(TypeCount++)
                   {
                     case 0: ROMTypeA=ROMTypeB=atoi(argv[N]);break;
                     case 1: ROMTypeB=atoi(argv[N]);break;
                     default: printf("%s: Excessive -rom option\n",argv[0]);
                   }
                 break;
        case 15: MSXVersion=0;break;
        case 16: MSXVersion=1;break;
        case 17: MSXVersion=2;break;
        case 18: N++;
                 if(N>=argc)
                   printf("%s: No joystick type supplied\n",argv[0]);
                 else
                   switch(JoyCount++)
                   {
                     case 0: JoyTypeA=atoi(argv[N]);break;
                     case 1: JoyTypeB=atoi(argv[N]);break;
                     default: printf("%s: Excessive -joy option\n",argv[0]);
                   }
                 break;
#ifdef DEBUG
        case 19: N++;
                 if(N>=argc)
                   printf("%s: No trap address supplied\n",argv[0]);
                 else
                   if(!strcmp(argv[N],"now")) CPU.Trace=1;
                   else sscanf(argv[N],"%hX",&(CPU.Trap));
                 break;
#endif
#ifdef SOUND
        case 20: UseSound=1;break;
        case 21: UseSound=0;break;
#endif
#ifdef UNIX
#ifdef MITSHM
        case 22: UseSHM=1;break;
        case 23: UseSHM=0;break;
#endif
        case 24: SaveCPU=1;break;
        case 25: SaveCPU=0;break;
        case 26: N++;
                 if(N<argc) SyncFreq=atoi(argv[N]);
                 else printf("%s: No sync frequency supplied\n",argv[0]);
                 break;
        case 27: SyncFreq=0;break;
        case 28: N++;
                 if(N<argc) UseZoom=atoi(argv[N]);
                 else printf("%s: No scaling factor supplied\n",argv[0]);
                 break;
#endif
        /* Wait VDP */
        case 29: UseVDPWait=1;break;
        case 30: UseVDPWait=0;break;
        /* Rensha */
        case 31: N++;
                 if(N<argc) RenshaWait=atoi(argv[N]);
                 else printf("%s: No rensha wait supplied\n",argv[0]);
                 break;
        case 32: RenshaWait=0;break;
        default: printf("%s: Wrong option '%s'\n",argv[0],argv[N]);
      }
    }

  /* Terminate disk lists and set initial disk names */
  if(DiskCount[0]) { Disks[0][DiskCount[0]]=0;DiskA=Disks[0][0]; }
  if(DiskCount[1]) { Disks[1][DiskCount[1]]=0;DiskB=Disks[1][0]; }

  /* Start fMSX! */
  if(!InitMachine()) return(1);
  StartMSX();
  TrashMSX();
  TrashMachine();
  return(0);
}
